import os
import sys

import pytest
from dotenv import load_dotenv

import crypto_connector as cc
from crypto_connector.base.exchange import Exchange

load_dotenv()


CREDENTIALS = {
    "Binance": {
        "sub_api_key": os.getenv("BINANCE_SUB_API_KEY_TEST"),
        "sub_api_secret": os.getenv("BINANCE_SUB_API_SECRET_TEST"),
        "sub_email": os.getenv("BINANCE_SUB_EMAIL_TEST"),
        "master_api_key": os.getenv("BINANCE_MASTER_API_KEY_TEST"),
        "master_api_secret": os.getenv("BINANCE_MASTER_API_SECRET_TEST"),
    },
    "HTX": {
        "api_key": os.getenv("HTX_API_KEY_TEST"),
        "api_secret": os.getenv("HTX_API_SECRET_TEST"),
    },
}


@pytest.fixture(scope="session")
def exchange(request: pytest.FixtureRequest) -> Exchange:
    # cannot run 'exchange' tests on github runners (because hosted in the US)
    if not sys.platform.startswith("win"):
        pytest.skip(
            reason="tests for windows only",
            allow_module_level=True,
        )

    exchange_name = request.param
    exchange = getattr(cc, exchange_name)(**CREDENTIALS[exchange_name])
    return exchange



import pytest

from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import Balance


@pytest.mark.parametrize(
    "exchange",
    argvalues=[
        ("Binance"),
        ("HTX"),
    ],
    indirect=["exchange"],
)
def test_get_api_key_info(exchange: Exchange) -> None:
    info = exchange.get_api_key_info()
    assert info["timestamp"] > 0


@pytest.mark.parametrize(
    "exchange",
    argvalues=[
        ("Binance"),
        ("HTX"),
    ],
    indirect=["exchange"],
)
def test_get_balance(exchange: Exchange) -> None:
    balance = exchange.get_balance()
    assert Balance.model_validate(balance)
    assert balance["equity"] >= 0


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import Transfer


@pytest.mark.parametrize(
    "exchange",
    argvalues=[
        ("Binance"),
        ("HTX"),
    ],
    indirect=["exchange"],
)
def test_get_transfer_history(exchange: Exchange) -> None:
    transfers = exchange.get_transfer_history()
    if not transfers:
        pytest.skip("No transfers found")
    assert [Transfer.model_validate(tr) for tr in transfers]


if __name__ == "__main__":
    pytest.main([__file__])



import hashlib
import hmac
import time
from datetime import datetime
from json import JSONDecodeError
from typing import Any, Literal
from urllib.parse import urlencode

from requests import Response

from crypto_connector.base.errors import (
    BadResponse,
    ExchangeError,
    MissingCredentials,
    OrderNotFound,
)
from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import (
    API,
    Balance,
    BalanceAsset,
    Market,
    Order,
    OrderCancelled,
    OrderStatus,
    TranferStatus,
    Transfer,
)


class Binance(Exchange):
    base_url = "https://api.binance.com"
    name = "Binance"
    order_statuses = {
        "NEW": OrderStatus.open,
        "PARTIALLY_FILLED": OrderStatus.open,
        "ACCEPTED": OrderStatus.open,
        "FILLED": OrderStatus.closed,
        "CANCELED": OrderStatus.canceled,
        "CANCELLED": OrderStatus.canceled,
        "PENDING_CANCEL": OrderStatus.canceling,
        "REJECTED": OrderStatus.rejected,
        "EXPIRED": OrderStatus.expired,
        "EXPIRED_IN_MATCH": OrderStatus.expired,
    }
    transfer_statuses = {
        "SUCCESS": TranferStatus.success,
        "PROCESS": TranferStatus.pending,
        "FAILURE": TranferStatus.failed,
    }
    timeframes = {
        "1m": "1m",
        "5m": "5m",
        "15m": "15m",
        "1h": "1h",
        "4h": "4h",
        "1d": "1d",
        "1w": "1w",
        "1M": "1M",
    }

    def __init__(
        self,
        sub_api_key: str | None = None,
        sub_api_secret: str | None = None,
        sub_email: str | None = None,
        master_api_key: str | None = None,
        master_api_secret: str | None = None,
        headers: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()
        self.sub_api_key = sub_api_key
        self.sub_api_secret = sub_api_secret
        self.sub_email = sub_email
        self.master_api_key = master_api_key
        self.master_api_secret = master_api_secret
        self.headers = headers

    @staticmethod
    def get_timestamp() -> int:
        return int(time.time() * 1000)

    @staticmethod
    def _cast_values(params: dict) -> dict:
        upper_params = ["symbol"]
        out = {}
        for key, value in params.items():
            if key in upper_params:
                if isinstance(key, str):
                    out[key] = value.upper()
                else:
                    raise ValueError(f"{key} should be a string.")
            else:
                if isinstance(value, bool):
                    out[key] = str(value).lower()
                else:
                    out[key] = str(value)
        return out

    @staticmethod
    def hash(api_secret: str, query_str: str) -> str:
        hash_obj = hmac.new(
            key=api_secret.encode("utf-8"),
            msg=query_str.encode("utf-8"),
            digestmod=hashlib.sha256,
        )
        return hash_obj.hexdigest()

    @staticmethod
    def _urlencode_dict(query: dict) -> str:
        return urlencode(query, doseq=True)

    def signed_request(
        self,
        http_method: str,
        url_path: str,
        params: dict | None = None,
        data: dict | None = None,
        master: bool = False,
    ) -> Any:
        if (
            (not self.sub_api_key)
            or (not self.sub_api_secret)
            or (not self.sub_email)
            or (not self.master_api_key)
            or (not self.master_api_secret)
        ):
            raise MissingCredentials(
                "To use private endpoints, user must pass credentials."
            )

        if params and data:
            raise ValueError("Can only pass `params` or `data`, but not both.")

        if master:
            api_key = self.master_api_key
            api_secret = self.master_api_secret
        else:
            api_key = self.sub_api_key
            api_secret = self.sub_api_secret

        payload = params or data or {}
        payload.update(
            {"timestamp": self.get_timestamp(), "recvWindow": self.recv_window}
        )
        payload = self._cast_values(self._clean_none_values(payload))
        urlencoded_payload = self._urlencode_dict(payload)
        signature = self.hash(api_secret, query_str=urlencoded_payload)  # type: ignore[arg-type] # noqa: E501
        payload.update({"signature": signature})
        self.session.headers.update({"X-MBX-APIKEY": api_key})
        if http_method == "GET":
            return self.request(http_method, url_path=url_path, params=payload)
        else:
            return self.request(http_method, url_path=url_path, data=payload)

    def handle_exception(self, r: Response) -> None:
        if r.status_code == 200:
            return

        try:
            rjson = r.json()
        except JSONDecodeError:
            raise BadResponse(f"Could not decode response text: {r.text}")

        error = {}
        error["error_code"] = rjson["code"]
        error["msg"] = rjson["msg"]
        error["url"] = r.url
        if self.headers:
            error["headers"] = r.headers
        raise ExchangeError(error)

    #################
    # EXCHANGE INFO #
    #################
    def get_server_time(self) -> int:
        """
        Fetch the current timestamp in millisecondsfrom the exchange server.
        :see: https://binance-docs.github.io/apidocs/spot/en/#test-connectivity
        """  # noqa: E501
        r = self.request("GET", "/api/v3/time")
        # {
        #   "serverTime": 1499827319559
        # }
        return r.get("serverTime", -1)

    ###########
    # ACCOUNT #
    ###########
    def _parse_api_key_info(self, api_info: dict[str, Any]) -> dict[str, Any]:
        api_obj = API(
            created=api_info["createTime"],
            timestamp=api_info["createTime"],
            spot_enabled=api_info["enableSpotAndMarginTrading"],
            ip_restricted=api_info["ipRestrict"],
            ips=api_info["ipList"],
            info=api_info,
        )
        return api_obj.model_dump()

    def get_api_key_info(self) -> dict[str, Any]:
        """
        Return API KEY info of the current user.
        :see: https://binance-docs.github.io/apidocs/spot/en/#get-api-key-permission-user_data
        """  # noqa: E501
        r = self.signed_request("GET", "/sapi/v1/account/apiRestrictions")
        # {
        #     'ipRestrict': True,
        #     'createTime': 1670109589000,
        #     'enableReading': True,
        #     'enableSpotAndMarginTrading': True,
        #     'enableWithdrawals': False,
        #     'enableInternalTransfer': False,
        #     'enableMargin': False,
        #     'enableFutures': False,
        #     'permitsUniversalTransfer': False,
        #     'enableVanillaOptions': False,
        #     'enablePortfolioMarginTrading': False
        # }
        params = {
            "email": self.sub_email,
            "subAccountApiKey": self.sub_api_key,
        }
        r2 = self.signed_request(
            "GET",
            "/sapi/v1/sub-account/subAccountApi/ipRestriction",
            params=params,
            master=True,
        )
        # {
        #     "ipRestrict": "true",
        #     "ipList": [
        #         "69.210.67.14",
        #         "8.34.21.10"
        #     ],
        #     "updateTime": 1636371437000,
        #     "apiKey": "XXXXX"
        # }
        r["ipList"] = r2["ipList"]
        api_key_info = self._parse_api_key_info(r)
        return api_key_info

    def _get_balance_value(self):
        """
        Get balance value in dollars.
        :see: https://binance-docs.github.io/apidocs/spot/en/#query-user-wallet-balance-user_data
        """  # noqa: E501
        r = self.signed_request("GET", "/sapi/v1/asset/wallet/balance")
        spot_balance_btc = float(
            [balance["balance"] for balance in r if balance["walletName"] == "Spot"][0]
        )
        balance_usd = spot_balance_btc * self.get_market_price("BTCUSDT")
        return balance_usd

    @staticmethod
    def _parse_balance_asset(asset: dict[str, Any]) -> BalanceAsset:
        return BalanceAsset(
            coin=asset["asset"],
            free=asset["free"],
            total=float(asset["free"]) + float(asset["locked"]),
        )

    def get_balance(self) -> dict:
        """
        Query for balance and get the amount of funds available for trading
        or funds locked in orders.
        :see: https://binance-docs.github.io/apidocs/spot/en/#account-information-user_data
        """  # noqa: E501
        params = {"omitZeroBalances": True}
        r = self.signed_request("GET", "/api/v3/account", params=params)
        # {'accountType': 'SPOT',
        # 'balances':
        #   [{'asset': 'BTC', 'free': '0.00000907', 'locked': '0.00000000'},
        #    {'asset': 'ETH', 'free': '0.00006900', 'locked': '0.00000000'},
        #    {'asset': 'USDT', 'free': '19.94000000', 'locked': '0.00000000'},
        #    {'asset': 'UNI', 'free': '0.99900000', 'locked': '0.00000000'}],
        # 'brokered': False,
        # 'buyerCommission': 0,
        # 'canDeposit': True,
        # 'canTrade': True,
        # 'canWithdraw': True,
        # 'commissionRates': {'buyer': '0.00000000',
        #                     'maker': '0.00100000',
        #                     'seller': '0.00000000',
        #                     'taker': '0.00100000'},
        # 'makerCommission': 10,
        # 'permissions': ['TRD_GRP_011'],
        # 'preventSor': False,
        # 'requireSelfTradePrevention': False,
        # 'sellerCommission': 0,
        # 'takerCommission': 10,
        # 'uid': 526318869,
        # 'updateTime': 1718115238667}
        balance_assets = []
        for raw_asset in r["balances"]:
            asset = self._parse_balance_asset(raw_asset)
            balance_assets.append(asset)
        balance_usd = self._get_balance_value()
        balance = Balance(equity=balance_usd, assets=balance_assets)
        return balance.model_dump()

    def _parse_transfer(self, transfer: dict[str, Any]) -> dict[str, Any]:
        tr = Transfer(
            transferId=transfer["tranId"],
            date=transfer["time"],
            status=self.transfer_statuses[transfer["status"]],
            from_id=transfer["from"],
            to_id=transfer["to"],
            direction="in" if transfer["to"] == self.sub_email else "out",
            coin=transfer["asset"],
            qty=transfer["qty"],
            info=transfer,
        )
        return tr.model_dump()

    def get_transfer_history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        **kwargs,
    ) -> list[dict]:
        """
        Return transfer history (in and out) of the current account.
        :see: https://binance-docs.github.io/apidocs/spot/en/#query-sub-account-spot-asset-transfer-history-for-master-account
        """  # noqa: E501
        params = {**kwargs}
        if start_date:
            params.update({"startTime": self.dt_to_unix(start_date)})
        if end_date:
            params.update({"endTime": self.dt_to_unix(end_date)})
        transfers = []
        for key in ["fromEmail", "toEmail"]:
            params.update({key: self.sub_email})
            tr = self.signed_request(
                "GET",
                "/sapi/v1/sub-account/sub/transfer/history",
                params=params,
                master=True,
            )
            transfers.extend(tr)
            params.pop(key)
        # [{'asset': 'USDT',
        # 'from': 'xxx@xxx.com',
        # 'qty': '30.00000000',
        # 'status': 'SUCCESS',
        # 'time': 1716820838000,
        # 'to': 'yyy@yyy.com',
        # 'tranId': 175189153797}]
        transfers = [self._parse_transfer(tr) for tr in transfers]
        return transfers

    ###########
    # MARKETS #
    ###########
    def _parse_market(self, market) -> dict[str, Any]:
        new_filters = {item["filterType"]: item for item in market["filters"]}

        market_obj = Market(
            name=market["symbol"],
            active=market["status"] == "TRADING",
            base=market["baseAsset"],
            info=market,
            min_amt=float(new_filters["NOTIONAL"]["minNotional"]),
            min_qty=float(new_filters["LOT_SIZE"]["minQty"]),
            precision=self.decimal_places(float(new_filters["LOT_SIZE"]["stepSize"])),
            quote=market["quoteAsset"],
            spot=market["isSpotTradingAllowed"],
        )
        return market_obj.model_dump()

    def get_markets(self, **kwargs) -> list[dict]:
        """
        Retrieve data on all SPOT markets.
        :see: https://binance-docs.github.io/apidocs/spot/en/#exchange-information
        """  # noqa: E501
        params = {"permissions": "SPOT", **kwargs}
        r = self.request("GET", "/api/v3/exchangeInfo", params=params)
        # {'active': True,
        # 'base': 'ETH',
        # 'info': {'allowTrailingStop': True,
        #         'allowedSelfTradePreventionModes': ['EXPIRE_TAKER',
        #                                             'EXPIRE_MAKER',
        #                                             'EXPIRE_BOTH'],
        #         'baseAsset': 'ETH',
        #         'baseAssetPrecision': 8,
        #         'baseCommissionPrecision': 8,
        #         'cancelReplaceAllowed': True,
        #         'defaultSelfTradePreventionMode': 'EXPIRE_MAKER',
        #         'filters': [{'filterType': 'PRICE_FILTER',
        #                     'maxPrice': '1000000.00000000',
        #                     'minPrice': '0.01000000',
        #                     'tickSize': '0.01000000'},
        #                     {'filterType': 'LOT_SIZE',
        #                     'maxQty': '9000.00000000',
        #                     'minQty': '0.00010000',
        #                     'stepSize': '0.00010000'},
        #                     {'filterType': 'ICEBERG_PARTS', 'limit': 10},
        #                     {'filterType': 'MARKET_LOT_SIZE',
        #                     'maxQty': '2040.56915833',
        #                     'minQty': '0.00000000',
        #                     'stepSize': '0.00000000'},
        #                     {'filterType': 'TRAILING_DELTA',
        #                     'maxTrailingAboveDelta': 2000,
        #                     'maxTrailingBelowDelta': 2000,
        #                     'minTrailingAboveDelta': 10,
        #                     'minTrailingBelowDelta': 10},
        #                     {'askMultiplierDown': '0.2',
        #                     'askMultiplierUp': '5',
        #                     'avgPriceMins': 5,
        #                     'bidMultiplierDown': '0.2',
        #                     'bidMultiplierUp': '5',
        #                     'filterType': 'PERCENT_PRICE_BY_SIDE'},
        #                     {'applyMaxToMarket': False,
        #                     'applyMinToMarket': True,
        #                     'avgPriceMins': 5,
        #                     'filterType': 'NOTIONAL',
        #                     'maxNotional': '9000000.00000000',
        #                     'minNotional': '5.00000000'},
        #                     {'filterType': 'MAX_NUM_ORDERS', 'maxNumOrders': 200},  # noqa: E501
        #                     {'filterType': 'MAX_NUM_ALGO_ORDERS',
        #                     'maxNumAlgoOrders': 5}],
        #         'icebergAllowed': True,
        #         'isMarginTradingAllowed': True,
        #         'isSpotTradingAllowed': True,
        #         'ocoAllowed': True,
        #         'orderTypes': ['LIMIT',
        #                         'LIMIT_MAKER',
        #                         'MARKET',
        #                         'STOP_LOSS_LIMIT',
        #                         'TAKE_PROFIT_LIMIT'],
        #         'otoAllowed': False,
        #         'permissionSets': [['SPOT',
        #                             'MARGIN',
        #                             'TRD_GRP_004',
        #                             'TRD_GRP_005',
        #                             'TRD_GRP_006',
        #                             'TRD_GRP_009',
        #                             'TRD_GRP_010',
        #                             'TRD_GRP_011',
        #                             'TRD_GRP_012',
        #                             'TRD_GRP_013',
        #                             'TRD_GRP_014',
        #                             'TRD_GRP_015',
        #                             'TRD_GRP_016',
        #                             'TRD_GRP_017',
        #                             'TRD_GRP_018',
        #                             'TRD_GRP_019',
        #                             'TRD_GRP_020',
        #                             'TRD_GRP_021',
        #                             'TRD_GRP_022',
        #                             'TRD_GRP_023',
        #                             'TRD_GRP_024',
        #                             'TRD_GRP_025']],
        #         'permissions': [],
        #         'quoteAsset': 'USDT',
        #         'quoteAssetPrecision': 8,
        #         'quoteCommissionPrecision': 8,
        #         'quoteOrderQtyMarketAllowed': True,
        #         'quotePrecision': 8,
        #         'status': 'TRADING',
        #         'symbol': 'ETHUSDT'},
        # 'min_amt': 0.01,
        # 'min_qty': 0.0001,
        # 'name': 'ETHUSDT',
        # 'precision': 4,
        # 'quote': 'USDT',
        # 'spot': True}
        markets = [self._parse_market(market) for market in r["symbols"]]
        return markets

    def get_market_info(self, market: str, **kwargs) -> dict[str, Any]:
        """
        Retrieve data on a specific market.
        :see: https://binance-docs.github.io/apidocs/spot/en/#exchange-information
        """  # noqa: E501
        if not market:
            raise ValueError(f"`market` cannot be empty, value passed: '{market}'")

        params = {"symbol": market, **kwargs}
        r = self.request("GET", "/api/v3/exchangeInfo", params=params)
        parsed_market: dict[str, Any] = self._parse_market(r["symbols"][0])
        return parsed_market

    @staticmethod
    def _parse_ohlcv(data: list[list]) -> list[list]:
        return [
            [
                int(x[0]),
                float(x[1]),
                float(x[2]),
                float(x[3]),
                float(x[4]),
                float(x[5]),
            ]
            for x in data
        ]

    def get_ohlcv(
        self,
        market: str,
        timeframe: str = "1d",
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        **kwargs,
    ) -> list[list]:
        """
        Fetch historical candlestick data containing
        the open, high, low, and close price, and the volume of a market.
        :see: https://binance-docs.github.io/apidocs/spot/en/#kline-candlestick-data
        """  # noqa: E501
        params = {
            "symbol": market,
            "interval": timeframe,
            "limit": kwargs.get("limit", 200),
            **kwargs,
        }
        if start_date:
            params.update({"startTime": self.dt_to_unix(start_date)})
        if end_date:
            params.update({"endTime": self.dt_to_unix(end_date)})

        r = self.request("GET", "/api/v3/klines", params=params)
        # [
        #   [
        #     1499040000000,      // Kline open time
        #     "0.01634790",       // Open price
        #     "0.80000000",       // High price
        #     "0.01575800",       // Low price
        #     "0.01577100",       // Close price
        #     "148976.11427815",  // Volume
        #     1499644799999,      // Kline Close time
        #     "2434.19055334",    // Quote asset volume
        #     308,                // Number of trades
        #     "1756.87402397",    // Taker buy base asset volume
        #     "28.46694368",      // Taker buy quote asset volume
        #     "0"                 // Unused field, ignore.
        #   ]
        # ]
        ohlcv = self._parse_ohlcv(r)
        return ohlcv

    def get_market_price(self, market: str) -> float:
        """
        Return the current price of a market.
        see: https://binance-docs.github.io/apidocs/spot/en/#symbol-price-ticker
        """  # noqa: E501
        params = {"symbol": market}
        r = self.request("GET", "/api/v3/ticker/price", params=params)
        # {"symbol": "LTCBTC", "price": "4.00000200"}
        return float(r["price"])

    #########
    # ORDER #
    #########
    def _parse_order(self, order: dict[str, Any]) -> dict[str, Any]:
        order_obj = Order(
            orderId=order["orderId"],
            dt=order["time"],
            market=order["symbol"],
            type=order["type"].lower(),
            side=order["side"].lower(),
            qty=order["origQty"],
            filled=order["executedQty"],
            amount=order["cummulativeQuoteQty"],
            price=order["price"],
            status=self.order_statuses[order["status"]],
            time_in_force=order["timeInForce"],
            fee=None,
            info=order,
        )
        return order_obj.model_dump()

    def get_order(self, id: str, market: str, **kwargs) -> dict[str, Any]:
        """
        Get information on an order made by the user.
        :see: https://binance-docs.github.io/apidocs/spot/en/#query-order-user_data
        """  # noqa: E501
        params = {"orderId": id, "symbol": market, **kwargs}
        r = self.signed_request("GET", "/api/v3/order", params=params)
        # {'clientOrderId': 'nOIUBuoybuyBUOYBbJnGfT',
        # 'cummulativeQuoteQty': '0.00000000',
        # 'executedQty': '0.00000000',
        # 'icebergQty': '0.00000000',
        # 'isWorking': True,
        # 'orderId': 17967599781,
        # 'orderListId': -1,
        # 'origQty': '0.01000000',
        # 'origQuoteOrderQty': '0.00000000',
        # 'price': '1000.00000000',
        # 'selfTradePreventionMode': 'EXPIRE_MAKER',
        # 'side': 'BUY',
        # 'status': 'NEW',
        # 'stopPrice': '0.00000000',
        # 'symbol': 'ETHUSDT',
        # 'time': 1717160522940,
        # 'timeInForce': 'GTC',
        # 'type': 'LIMIT',
        # 'updateTime': 1717160522940,
        # 'workingTime': 1717160522940}
        order = self._parse_order(r)
        return order

    def place_order(
        self,
        market: str,
        type: Literal["limit", "market"],
        side: Literal["buy", "sell"],
        qty: float,
        price: float | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        """
        Place an order.
        :see: https://binance-docs.github.io/apidocs/spot/en/#new-order-trade
        """
        time_in_force = None
        if type.lower() == "limit":
            time_in_force = kwargs.get("timeInForce", "GTC")

        data = {
            "symbol": market,
            "side": side,
            "type": type,
            "quantity": qty,
            "price": price,
            "timeInForce": time_in_force,
            **kwargs,
        }
        data.update(kwargs)
        r = self.signed_request("POST", "/api/v3/order", data=data)
        # {'clientOrderId': 'YUBKHJgbUYGVBYUGFVByuj',
        #  'cummulativeQuoteQty': '0.00000000',
        #  'executedQty': '0.00000000',
        #  'fills': [],
        #  'orderId': 17967599781,
        #  'orderListId': -1,
        #  'origQty': '0.01000000',
        #  'price': '1000.00000000',
        #  'selfTradePreventionMode': 'EXPIRE_MAKER',
        #  'side': 'BUY',
        #  'status': 'NEW',
        #  'symbol': 'ETHUSDT',
        #  'timeInForce': 'GTC',
        #  'transactTime': 1717160522940,
        #  'type': 'LIMIT',
        #  'workingTime': 1717160522940}
        order = self.get_order(id=r["orderId"], market=r["symbol"])
        return order

    def get_open_orders(
        self, market: str | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        """
        Get all currently unfilled open orders.
        :see: https://binance-docs.github.io/apidocs/spot/en/#current-open-orders-user_data
        """  # noqa: E501
        params = {"symbol": market, **kwargs}
        r = self.signed_request("GET", "/api/v3/openOrders", params=params)
        # [{'clientOrderId': 'NIOUniunui45BNIUjhbjHg',
        #   'cummulativeQuoteQty': '0.00000000',
        #   'executedQty': '0.00000000',
        #   'icebergQty': '0.00000000',
        #   'isWorking': True,
        #   'orderId': 17964589418,
        #   'orderListId': -1,
        #   'origQty': '0.01000000',
        #   'origQuoteOrderQty': '0.00000000',
        #   'price': '1000.00000000',
        #   'selfTradePreventionMode': 'EXPIRE_MAKER',
        #   'side': 'BUY',
        #   'status': 'NEW',
        #   'stopPrice': '0.00000000',
        #   'symbol': 'ETHUSDT',
        #   'time': 1717145002434,
        #   'timeInForce': 'GTC',
        #   'type': 'LIMIT',
        #   'updateTime': 1717145002434,
        #   'workingTime': 1717145002434}]
        orders = [self._parse_order(order) for order in r]
        return orders

    def cancel_order(self, id: str, **kwargs) -> dict[str, Any]:
        """
        Cancel an open order.
        :see: https://binance-docs.github.io/apidocs/spot/en/#cancel-order-trade
        """  # noqa: E501
        orders = self.get_open_orders()
        market = [order["market"] for order in orders if order["orderId"] == id]
        if not market:
            raise OrderNotFound(f"Could not find an open order with this id '{id}'")

        data = {"orderId": id, "symbol": market[0], **kwargs}
        r = self.signed_request("DELETE", "/api/v3/order", data=data)
        # {'clientOrderId': 'niluBUYbuykvbUYLB156NU',
        #  'cummulativeQuoteQty': '0.00000000',
        #  'executedQty': '0.00000000',
        #  'orderId': 18007366952,
        #  'orderListId': -1,
        #  'origClientOrderId': 'bIULYGBjkhb6K7GKuyvuyv',
        #  'origQty': '0.01000000',
        #  'price': '1000.00000000',
        #  'selfTradePreventionMode': 'EXPIRE_MAKER',
        #  'side': 'BUY',
        #  'status': 'CANCELED',
        #  'symbol': 'ETHUSDT',
        #  'timeInForce': 'GTC',
        #  'transactTime': 1717505772211,
        #  'type': 'LIMIT'}
        order_obj = OrderCancelled(
            orderId=r["orderId"], success=(r["status"] == "CANCELED")
        )
        return order_obj.model_dump()

    def cancel_orders(
        self, market: str | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        """
        Cancel all open orders.
        :see: https://binance-docs.github.io/apidocs/spot/en/#cancel-all-open-orders-on-a-symbol-trade
        """  # noqa: E501
        orders = self.get_open_orders()
        markets = [order["market"] for order in orders]
        resp = []
        data: dict[str, Any] = {}
        if market:
            if market not in markets:
                return []
            data = {"symbol": market}
            resp = self.signed_request("DELETE", "/api/v3/openOrders", data=data)
        else:
            for market in markets:
                data = {"symbol": market}
                r = self.signed_request("DELETE", "/api/v3/openOrders", data=data)
                resp.extend(r)

        cancelled_orders = []
        for order in resp:
            cancelled_orders.append(
                OrderCancelled(
                    orderId=order["orderId"], success=order["status"] == "CANCELED"
                ).model_dump()
            )
        return cancelled_orders



import pytest

from crypto_connector.base.errors import ExchangeError
from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import Market


@pytest.mark.parametrize(
    "exchange,market",
    argvalues=[
        ("Binance", "SOLUSDT"),
        ("Binance", "solusdt"),
        ("HTX", "SOLUSDT"),
        ("HTX", "solusdt"),
    ],
    indirect=["exchange"],
)
def test_get_market_info(exchange: Exchange, market: str) -> None:
    info = exchange.get_market_info(market)
    assert isinstance(Market.model_validate(info), Market)


@pytest.mark.parametrize(
    "exchange,market",
    argvalues=[
        ("Binance", "XYZ"),
        ("HTX", "XYZ"),
    ],
    indirect=["exchange"],
)
def test_get_market_info_wrong_symbol(exchange: Exchange, market: str) -> None:
    with pytest.raises(ExchangeError):
        exchange.get_market_info(market)


@pytest.mark.parametrize(
    "exchange,market",
    argvalues=[
        ("Binance", ""),
        ("HTX", ""),
    ],
    indirect=["exchange"],
)
def test_get_market_info_empty_symbol(exchange: Exchange, market: str) -> None:
    with pytest.raises(ValueError):
        exchange.get_market_info(market)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from crypto_connector.base.exchange import Exchange


@pytest.mark.parametrize(
    "exchange,market",
    argvalues=[
        ("Binance", "XRPUSDC"),
        ("HTX", "XRPUSDT"),
    ],
    indirect=["exchange"],
)
def test_get_market_price(exchange: Exchange, market: str) -> None:
    assert exchange.get_market_price(market) > 0


if __name__ == "__main__":
    pytest.main([__file__])



import base64
import hashlib
import hmac
import json
from datetime import datetime, timezone
from json.decoder import JSONDecodeError
from typing import Any, Literal
from urllib.parse import urlencode, urlparse

from requests.models import Response

from crypto_connector.base.errors import BadResponse, ExchangeError, MissingCredentials
from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import (
    API,
    Balance,
    BalanceAsset,
    Market,
    Order,
    OrderCancelled,
    OrderStatus,
    TranferStatus,
    Transfer,
)


class HTX(Exchange):
    base_url = "https://api.huobi.pro"  # or api-aws.huobi.pro for aws client
    name = "HTX"
    order_statuses = {
        # spot
        "partial-filled": OrderStatus.open,
        "partial-canceled": OrderStatus.canceled,
        "filled": OrderStatus.closed,
        "canceled": OrderStatus.canceled,
        "submitted": OrderStatus.open,
        "created": OrderStatus.open,  # For stop orders
    }
    timeframes = {
        "1m": "1min",
        "5m": "5min",
        "15m": "15min",
        "1h": "60min",
        "4h": "4hour",
        "1d": "1day",
        "1w": "1week",
        "1M": "1mon",
    }

    def __init__(
        self,
        api_key: str | None = None,
        api_secret: str | None = None,
        headers: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()
        self.api_key = api_key
        self.api_secret = api_secret
        self.headers = headers

    @staticmethod
    def get_timestamp() -> str:
        return datetime.now(tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def _urlencode_dict(query: dict) -> str:
        return urlencode(query, True)

    @staticmethod
    def _cast_values(params: dict) -> dict:
        lower_params = ["symbol", "symbols"]
        out = {}
        for key, value in params.items():
            if key in lower_params:
                if isinstance(key, str):
                    out[key] = value.lower()
                else:
                    raise ValueError(f"{key} should be a string.")
            else:
                out[key] = str(value)
        return out

    def _urlencode_params(self, params: dict) -> str:
        return self._urlencode_dict(self._cast_values(self._clean_none_values(params)))

    def _prepare_request_data(self, data: dict) -> str:
        casted_data = self._cast_values(self._clean_none_values(data))
        return json.dumps(casted_data)

    @staticmethod
    def sort_dict(d: dict) -> dict:
        """ASCII sort dict"""
        if not isinstance(d, dict):
            raise ValueError("The input must be a dictionary")
        return dict(sorted(d.items()))

    @staticmethod
    def hash(api_secret: str, query_str: str) -> bytes:
        hash_obj = hmac.new(
            api_secret.encode("utf-8"),
            msg=query_str.encode("utf-8"),
            digestmod=hashlib.sha256,
        )
        return hash_obj.digest()

    @staticmethod
    def base64_encode(hash: bytes) -> str:
        return base64.b64encode(hash).decode()

    def signed_request(
        self,
        http_method: str,
        url_path: str,
        params: dict | None = None,
        data: dict | None = None,
    ) -> dict[str, Any]:
        if (not self.api_key) or (not self.api_secret):
            raise MissingCredentials(
                "To use private endpoints, user must pass credentials."
            )

        if params and data:
            raise ValueError("Can only pass `params` or `data`, but not both.")

        payload = params or data or {}
        to_sign_payload = {
            "AccessKeyId": self.api_key,
            "SignatureMethod": "HmacSHA256",
            "SignatureVersion": 2,
            "Timestamp": self.get_timestamp(),
        }
        if http_method == "GET":
            to_sign_payload.update(payload)

        encoded_payload = self._urlencode_params(self.sort_dict(to_sign_payload))
        host = urlparse(self.base_url).hostname
        to_hash = f"{http_method}\n{host}\n{url_path}\n{encoded_payload}"
        hash_obj = self.hash(self.api_secret, query_str=to_hash)  # type: ignore[arg-type]  # noqa: E501
        signature = self.base64_encode(hash_obj)
        to_sign_payload.update({"Signature": signature})
        url_path += "?" + self._urlencode_params(to_sign_payload)

        if http_method == "GET":
            return self.request(http_method, url_path=url_path, params=None)
        else:
            prep_data = self._prepare_request_data(data=payload)
            return self.request(http_method, url_path=url_path, data=prep_data)

    def handle_exception(self, r: Response) -> None:
        try:
            rjson = r.json()
        except JSONDecodeError:
            raise BadResponse(f"Could not decode response text: {r.text}")

        if rjson.get("status") == "ok" or rjson.get("ok") is True:
            return

        error = {}
        error["error_code"] = rjson.get("err-code") or rjson.get("code")
        error["msg"] = rjson.get("err-msg") or rjson.get("message")
        error["url"] = r.url
        if self.headers:
            error["headers"] = r.headers
        raise ExchangeError(error)

    ###############################
    # EXCHANGE SPECIFIC ENDPOINTS #
    ###############################
    def _get_uid(self) -> int:
        """This endpoint allow users to view the user ID of the account easily.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec52d6c-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        r = self.signed_request("GET", "/v2/user/uid")
        # {'code': 200, 'data': 000000000, 'ok': True}
        return r.get("data", -1)

    @property
    def _account_id(self) -> int:
        r = self.signed_request("GET", "/v1/account/accounts")
        # {
        # "data":[
        #     {
        #         "id":01234567,
        #         "state":"working",
        #         "subtype":"",
        #         "type":"spot"
        #     }
        # ],
        # "status":"ok"
        # }
        return r.get("data", -1)[0]["id"]

    #################
    # EXCHANGE INFO #
    #################
    def get_server_time(self) -> int:
        """
        Fetch the current timestamp in millisecondsfrom the exchange server
        :see: https://www.htx.com/en-in/opend/newApiPages/?id=7ec4bb2c-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        r = self.request("GET", "/v1/common/timestamp")
        # {"status": "ok", "data": 1629715504949}
        return r.get("data", -1)

    # ###########
    # # ACCOUNT #
    # ###########
    def _parse_api_key_info(self, api_info: dict[str, Any]) -> dict[str, Any]:
        api_obj = API(
            created=api_info["createTime"],
            timestamp=api_info["createTime"],
            spot_enabled="trade" in api_info["permission"],
            ip_restricted=len(api_info["ipAddresses"]) > 0,
            ips=api_info["ipAddresses"].split(","),
            info=api_info,
        )
        return api_obj.model_dump()

    def get_api_key_info(self) -> dict[str, Any]:
        """
        Return API KEY info of the current user.
        :see: https://www.htx.com/en-in/opend/newApiPages/?id=7ec52c92-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        params = {"uid": self._get_uid(), "accessKey": self.api_key}
        r = self.signed_request("GET", "/v2/user/api-key", params=params)
        # {
        #     'code': 200,
        #     'data': [{'accessKey': '5cd012ef-ez2xc4vb6n-29e57c70-72b72',
        #            'createTime': 1716580080000,
        #            'ipAddresses': 'XX.XX.XX.XXX,XX.XXX.XXX.XX',
        #            'note': 'trading bot api',
        #            'permission': 'readOnly,trade',
        #            'status': 'normal',
        #            'updateTime': 1716580080000,
        #            'validDays': -1}],
        #     'message': 'success',
        #     'ok': True
        # }
        data = r.get("data")
        if not data:
            raise BadResponse(f"No data returned from server {r=}.")

        api_key_info = self._parse_api_key_info(data[0])
        return api_key_info

    def _get_balance_value(self):
        """Get balance value in dollars.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4ff6d-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        params = {"accountType": "spot", "valuationCurrency": "USD"}
        r = self.signed_request("GET", "/v2/account/asset-valuation", params=params)
        balance_usd = r["data"]["balance"]
        return float(balance_usd)

    @staticmethod
    def _parse_balance_asset(asset: dict) -> BalanceAsset:
        return BalanceAsset(
            coin=asset["currency"],
            free=asset["available"],
            total=asset["available"],
        )

    def get_balance(self) -> dict:
        """
        Query for balance and get the amount of funds available for trading
        or funds locked in orders.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4b429-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        account_id = self._account_id
        r = self.signed_request("GET", f"/v1/account/accounts/{account_id}/balance")
        # {
        #   "status": "ok",
        #   "data": {
        #     "id": 1000001,
        #     "type": "spot",
        #     "state": "working",
        #     "list": [
        #       {
        #         "currency": "usdt",
        #         "type": "trade",
        #         "balance": "91.850043797676510303",
        #         "debt": "invalid",
        #         "available": "invalid",
        #         "seq-num": "477"
        #       },
        #       {
        #         "currency": "usdt",
        #         "type": "frozen",
        #         "balance": "5.160000000000000015",
        #         "debt": "invalid",
        #         "available": "invalid",
        #         "seq-num": "477"
        #       },
        #       {
        #         "currency": "poly",
        #         "type": "trade",
        #         "balance": "147.928994082840236",
        #         "debt": "invalid",
        #         "available": "invalid",
        #         "seq-num": "2"
        #       }
        #     ]
        #   }
        # }
        balance_assets = []
        for raw_asset in r["data"]["list"]:
            if raw_asset["type"] == "frozen":
                continue
            asset = self._parse_balance_asset(raw_asset)
            if asset.free == 0:
                continue
            balance_assets.append(asset)
        balance_usd = self._get_balance_value()
        balance = Balance(equity=balance_usd, assets=balance_assets)
        return balance.model_dump()

    def _parse_transfer(self, transfer: dict[str, Any]) -> dict[str, Any]:
        amt = float(transfer["transact-amt"])
        tr = Transfer(
            transferId=transfer["record-id"],
            date=transfer["transact-time"],
            status=TranferStatus.success,
            from_id=transfer["source-id"],
            to_id=transfer["account-id"],
            direction=("in" if amt > 0 else "out"),
            coin=transfer["currency"],
            qty=abs(amt),
            info=transfer,
        )
        return tr.model_dump()

    def get_transfer_history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        **kwargs,
    ) -> list[dict]:
        """
        Return transfer history (in and out) of the current account.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4b85b-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        params = {
            "account-id": self._account_id,
            "transact-types": "transfer",
            **kwargs,
        }
        if start_date:
            params.update({"start-time": self.dt_to_unix(start_date)})
        if end_date:
            params.update({"end-time": self.dt_to_unix(end_date)})
        r = self.signed_request("GET", "/v1/account/history", params=params)
        # {'data': [{'account-id': 61775802,
        #         'acct-balance': '30.0000000000000000',
        #         'avail-balance': '30.0000000000000000',
        #         'currency': 'usdt',
        #         'record-id': 1104073401938149557,
        #         'source-id': 3068183988,
        #         'transact-amt': '30.0000000000000000',
        #         'transact-time': 1716821954841,
        #         'transact-type': 'transfer'}],
        # 'status': 'ok'}
        transfers = [self._parse_transfer(tr) for tr in r["data"]]
        return transfers

    # ###########
    # # MARKETS #
    # ###########
    def _parse_market(self, market) -> dict[str, Any]:
        market_obj = Market(
            name=market["symbol"],
            active=market["state"] == "online",
            base=market["bc"],
            info=market,
            min_amt=market["minov"],
            min_qty=market["minoa"],
            precision=market["ap"],
            quote=market["qc"],
            spot=True,
        )
        market = market_obj.model_dump()
        return market

    def get_markets(self, **kwargs) -> list[dict]:
        """
        Retrieves data on all (SPOT/MARGIN) markets.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4f5d6-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        params = {**kwargs}
        r = self.request("GET", "/v1/settings/common/market-symbols", params=params)
        # {
        # "status": "ok",
        # "data": [
        #     {
        #     "symbol": "btc3lusdt",
        #     "state": "online",
        #     "bc": "btc3l",
        #     "qc": "usdt",
        #     "pp": 4,
        #     "ap": 4,
        #     "sp": "main",
        #     "vp": 8,
        #     "minoa": 0.01,
        #     "maxoa": 199.0515,
        #     "minov": 5,
        #     "lominoa": 0.01,
        #     "lomaxoa": 199.0515,
        #     "lomaxba": 199.0515,
        #     "lomaxsa": 199.0515,
        #     "smminoa": 0.01,
        #     "blmlt": 1.1,
        #     "slmgt": 0.9,
        #     "smmaxoa": 199.0515,
        #     "bmmaxov": 2500,
        #     "msormlt": 0.1,
        #     "mbormlt": 0.1,
        #     "maxov": 2500,
        #     "u": "btcusdt",
        #     "mfr": 0.035,
        #     "ct": "23:55:00",
        #     "rt": "00:00:00",
        #     "rthr": 4,
        #     "in": 16.3568,
        #     "at": "enabled",
        #     "tags": "etp,nav,holdinglimit,activities"
        #     }
        # ],
        # "ts": "1641880897191",
        # "full": 1
        # }
        markets = [self._parse_market(market) for market in r["data"]]
        return markets

    def get_market_info(self, market: str, **kwargs) -> dict[str, Any]:
        """
        Retrieves data on a specific market.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4f5d6-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        if not market:
            raise ValueError(f"`market` cannot be empty, value passed: '{market}'")

        params = {"symbols": market}
        r = self.request("GET", "/v1/settings/common/market-symbols", params=params)
        # {'data': [{'ap': 4,
        #            'at': 'enabled',
        #            'bc': 'eth',
        #            'blmlt': 1.1,
        #            'bmmaxov': 1000000,
        #            'lomaxba': 10000,
        #            'lomaxoa': 10000,
        #            'lomaxsa': 10000,
        #            'lominoa': 0.001,
        #            'lr': 5,
        #            'maxoa': 10000,
        #            'mbormlt': 0.1,
        #            'minoa': 0.001,
        #            'minov': 10,
        #            'msormlt': 0.1,
        #            'pp': 2,
        #            'qc': 'usdt',
        #            'slmgt': 0.9,
        #            'smlr': 3,
        #            'smmaxoa': 500,
        #            'smminoa': 0.001,
        #            'sp': 'main',
        #            'state': 'online',
        #            'symbol': 'ethusdt',
        #            'tags': '',
        #            'vp': 8}],
        #  'full': 1,
        #  'status': 'ok',
        #  'ts': '1717592823163'}
        parsed_market = self._parse_market(r["data"][0])
        return parsed_market

    def get_market_price(self, market: str) -> float:
        """
        Return the current price of a market.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4a2cd-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        params = {"symbol": market}
        r = self.request("GET", "/market/detail", params=params)
        # {'ch': 'market.btcusdt.detail',
        #  'status': 'ok',
        #  'tick': {'amount': 4482.995312047257,
        #           'close': 67956.81,
        #           'count': 1252252,
        #           'high': 68921.56,
        #           'id': 342212175986,
        #           'low': 67273.65,
        #           'open': 68122.0,
        #           'version': 342212175986,
        #           'vol': 305856161.36989886},
        #  'ts': 1716988289671}
        return r["tick"]["close"]

    # #########
    # # ORDER #
    # #########
    def _parse_order(self, order: dict[str, Any]) -> dict[str, Any]:
        # handle typos in HTX response
        filled_amount = order.get("filled-amount", order.get("field-amount", 0))
        filled_cash_amount = order.get(
            "filled-cash-amount", order.get("field-cash-amount", 0)
        )
        filled_fees = order.get("filled-fees", order.get("field-fees", 0))

        order_obj = Order(
            orderId=order["id"],
            dt=order["created-at"],
            market=order["symbol"],
            type=order["type"].split("-")[-1],
            side=order["type"].split("-")[0],
            qty=filled_amount,
            filled=filled_amount,
            amount=filled_cash_amount,
            price=order["price"],
            status=self.order_statuses[order["state"]],
            time_in_force=None,
            fee=filled_fees,
            info=order,
        )
        return order_obj.model_dump()

    def place_order(
        self,
        market: str,
        type: Literal["limit", "market"],
        side: Literal["buy", "sell"],
        qty: float,
        price: float | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        """
        Place an order.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4ee16-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        # qty: for buy market order, it's order value
        if (type.lower() == "market") and (side.lower() == "buy"):
            qty = self.compute_quote_qty(qty, market=market)

        data = {
            "account-id": self._account_id,
            "symbol": market,
            "type": f"{side}-{type}",
            "amount": qty,
            "price": price,
            **kwargs,
        }
        r = self.signed_request(
            http_method="POST",
            url_path="/v1/order/orders/place",
            data=data,
        )
        # {'data': '1086175400829771', 'status': 'ok'}
        order = self.get_order(r["data"])
        if order["price"] == 0:
            order["price"] = self.get_market_price(order["market"])
        return order

    # def _get_order_history(self, **kwargs) -> dict:
    #     """Query order history."""
    #     orders = []
    #     next_page_cursor = ""
    #     while True:
    #         r = self.client.get_order_history(
    #             category="spot", cursor=next_page_cursor, **kwargs
    #         )
    #         orders.extend(r["result"]["list"])
    #         next_page_cursor = r["result"]["nextPageCursor"]
    #         if not next_page_cursor:
    #             break
    #     return orders

    def get_order(self, id: str) -> dict:
        """Get information on an order made by the user.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4e31c-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        r = self.signed_request("GET", f"/v1/order/orders/{id}")
        # {'data': {'account-id': 61775802,
        #           'amount': '0.01',
        #           'canceled-at': 0,
        #           'client-order-id': '',
        #           'created-at': 1717085707510,
        #           'field-amount': '0',
        #           'field-cash-amount': '0',
        #           'field-fees': '0',
        #           'finished-at': 0,
        #           'ice-amount': '0',
        #           'id': 1077577805408812,
        #           'is-ice': False,
        #           'market-amount': '0',
        #           'price': '1000',
        #           'source': 'spot-api',
        #           'state': 'submitted',
        #           'symbol': 'ethusdt',
        #           'type': 'buy-limit',
        #           'updated-at': 1717085707510},
        #  'status': 'ok'}
        order = self._parse_order(r["data"])
        return order

    def get_open_orders(self, market: str | None = None, **kwargs) -> list[dict]:
        """
        Get all currently unfilled open orders.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4e04b-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        params = {"symbol": market, **kwargs}
        r = self.signed_request(
            http_method="GET", url_path="/v1/order/openOrders", params=params
        )
        # {'data': [{'account-id': 61775802,
        #         'amount': '0.010000000000000000',
        #         'client-order-id': '',
        #         'created-at': 1717085707510,
        #         'filled-amount': '0.0',
        #         'filled-cash-amount': '0.0',
        #         'filled-fees': '0.0',
        #         'ice-amount': '0.0',
        #         'id': 1077577805408812,
        #         'price': '1000.000000000000000000',
        #         'source': 'api',
        #         'state': 'submitted',
        #         'symbol': 'ethusdt',
        #         'type': 'buy-limit'}],
        # 'status': 'ok'}
        orders = [self._parse_order(order) for order in r["data"]]
        return orders

    def cancel_order(self, id: str) -> dict[str, Any]:
        """
        Cancels an open order.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4e938-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        r = self.signed_request("POST", f"/v1/order/orders/{id}/submitcancel")
        # {'data': '1077586051442359', 'status': 'ok'}
        order_obj = OrderCancelled(orderId=r["data"], success=(r["status"] == "ok"))
        return order_obj.model_dump()

    def cancel_orders(
        self, market: str | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        """
        Cancel all open orders.
        :see: https://www.htx.com/en-us/opend/newApiPages/?id=7ec4eb66-7773-11ed-9966-0242ac110003
        """  # noqa: E501
        orders = self.get_open_orders(market=market)
        if not orders:
            return []

        data = {"symbol": market, **kwargs}
        self.signed_request("POST", "/v1/order/orders/batchCancelOpenOrders", data=data)
        # {
        # "data":{
        #     "failed-count":0,
        #     "next-id":-1,
        #     "success-count":2
        # },
        # "status":"ok"
        # }
        cancelled_orders = []
        for order in orders:
            cancelled_orders.append(
                OrderCancelled(orderId=order["orderId"], success=True).model_dump()
            )
        return cancelled_orders



import pytest

from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import Market


@pytest.mark.parametrize(
    "exchange",
    argvalues=[
        ("Binance"),
        ("HTX"),
    ],
    indirect=["exchange"],
)
def test_get_markets(exchange: Exchange) -> None:
    markets = exchange.get_markets()
    for market in markets[:10]:
        assert isinstance(Market.model_validate(market), Market)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import Order


@pytest.mark.parametrize(
    "exchange",
    argvalues=[
        ("Binance"),
        ("HTX"),
    ],
    indirect=["exchange"],
)
def test_get_open_orders(exchange: Exchange) -> None:
    orders = exchange.get_open_orders()
    for order in orders:
        assert isinstance(Order.model_validate(order), Order)


if __name__ == "__main__":
    pytest.main([__file__])



from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import BaseModel, ConfigDict, StringConstraints


class API(BaseModel):
    created: datetime
    timestamp: int
    spot_enabled: bool
    ip_restricted: bool
    ips: list[str] | None = []
    info: dict[str, Any]


class Market(BaseModel):
    name: Annotated[str, StringConstraints(to_upper=True)]
    active: bool
    base: Annotated[str, StringConstraints(to_upper=True)]
    info: dict[str, Any]
    min_amt: float
    min_qty: float
    precision: int
    quote: Annotated[str, StringConstraints(to_upper=True)]
    spot: bool


# class Fee(BaseModel):
#     currency: str
#     cost: float
#     rate: float | None


class OrderStatus(str, Enum):
    open = "open"
    closed = "closed"
    canceled = "canceled"
    rejected = "rejected"
    canceling = "canceling"
    expired = "expired"


class Order(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True, use_enum_values=True)

    orderId: str
    dt: datetime
    market: Annotated[str, StringConstraints(to_upper=True)]
    type: Literal["limit", "market"]
    side: Literal["buy", "sell"]
    qty: float
    amount: float
    price: float
    status: OrderStatus
    time_in_force: str | None = None
    filled: float
    fee: float | None = None
    info: dict[str, Any]


class OrderCancelled(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)

    orderId: str
    success: bool


class BalanceAsset(BaseModel):
    coin: Annotated[str, StringConstraints(to_upper=True)]
    free: float
    total: float


class Balance(BaseModel):
    equity: float
    assets: list[BalanceAsset]


class TranferStatus(str, Enum):
    success = "success"
    pending = "pending"
    failed = "failed"


class Transfer(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True, use_enum_values=True)

    transferId: str
    date: datetime
    status: TranferStatus
    from_id: str
    to_id: str
    direction: Literal["in", "out"]
    coin: Annotated[str, StringConstraints(to_upper=True)]
    qty: float
    info: dict[str, Any]



import math
from datetime import datetime, timezone

import pytest

from crypto_connector.base.exchange import Exchange

exc = Exchange()


@pytest.mark.parametrize(
    argnames="f,decimal_places,expected",
    argvalues=[
        (10.123, 2, 10.12),
        (0.2999999999999997637, 2, 0.29),
        (10.0, 1, 10),
        (10.0, 1, 10),
        (10.0, 1, 10.0),
        (10.12323, 0, 10),
        (0.3, 1, 0.3),
        (0.3, 2, 0.3),
        (10e3, 1, 10000),
    ],
)
def test_truncate_float(f, decimal_places, expected) -> None:
    assert math.isclose(a=exc.truncate_float(f, decimal_places), b=expected, abs_tol=0)


@pytest.mark.parametrize(
    argnames="n,expected",
    argvalues=[
        (0.1, 1),
        (0.01, 2),
        (0.001, 3),
        (0.000001, 6),
        (0.0000001, 7),
        (0, 0),
        (1, 0),
        (1.0, 1),
        (-1, 0),
        (-0.01, 2),
    ],
)
def test_decimal_places(n, expected) -> None:
    assert exc.decimal_places(n) == expected


def test_decimal_places_wrong_input_type() -> None:
    with pytest.raises(ValueError):
        exc.decimal_places("0.01")  # type: ignore


@pytest.mark.parametrize(
    argnames="string,format,expected",
    argvalues=[
        (
            "2019-05-18T15:17:08Z",
            "%Y-%m-%dT%H:%M:%SZ",
            datetime(2019, 5, 18, 15, 17, 8, tzinfo=timezone.utc),
        ),
        (
            "2019-05-18T15:17:00+00:00",
            "%Y-%m-%dT%H:%M:%S%z",
            datetime(2019, 5, 18, 15, 17, tzinfo=timezone.utc),
        ),
    ],
)
def test_str_to_dt(string, format, expected) -> None:
    assert exc.str_to_dt(string=string, format=format) == expected


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from crypto_connector.base.errors import NotSupported
from crypto_connector.base.exchange import Exchange


@pytest.mark.parametrize(
    "exchange,market",
    argvalues=[
        ("Binance", "ETHUSDC"),
    ],
    indirect=["exchange"],
)
def test_get_ohlcv_ok(exchange: Exchange, market: str) -> None:
    ohlcv = exchange.get_ohlcv(market)
    assert len(ohlcv) > 0


@pytest.mark.parametrize(
    "exchange,market",
    argvalues=[
        ("HTX", "ETHUSDC"),
    ],
    indirect=["exchange"],
)
def test_get_ohlcv_not_supported(exchange: Exchange, market: str):
    with pytest.raises(NotSupported):
        exchange.get_ohlcv(market)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from crypto_connector.base.exchange import Exchange


@pytest.mark.parametrize(
    "exchange",
    argvalues=[
        ("Binance"),
        ("HTX"),
    ],
    indirect=["exchange"],
)
def test_get_server_time(exchange: Exchange) -> None:
    assert exchange.get_server_time() > 1


if __name__ == "__main__":
    pytest.main([__file__])
    pytest.main([__file__])



import logging
from datetime import datetime, timezone
from decimal import Decimal
from json.decoder import JSONDecodeError
from typing import Any, Literal

import requests

from crypto_connector.base.errors import NotSupported


class Exchange:
    """Base exchange class"""

    base_url: str
    name: str
    recv_window = 5000
    timeframes: dict[str, Any]
    timeout = 10000

    def __init__(self, **kwargs):
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Content-Type": "application/json",
                "Accept": "application/json",
            }
        )
        self.logger = logging.getLogger(__name__)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @staticmethod
    def _convert_to_str(d: dict) -> dict:
        out = {}
        for k in d.keys():
            if isinstance(d[k], bool):
                out[k] = str(d[k]).lower()
            elif isinstance(d[k], int | float):
                out[k] = str(d[k])
            else:
                out[k] = d[k]
        return out

    @staticmethod
    def _clean_none_values(d: dict) -> dict:
        out = {}
        for k in d.keys():
            if d[k] is not None:
                out[k] = d[k]
        return out

    @staticmethod
    def truncate_float(f: float, decimal_places: int) -> float | int:
        """Truncates a float f to n decimal places without rounding.
        Inspired by https://stackoverflow.com/a/783927
        """
        if not isinstance(f, int | float):
            raise ValueError("The input must be a float or an int.")

        if (not isinstance(decimal_places, int)) or (decimal_places < 0):
            raise ValueError("`decimal_places` argument must a positive int.")

        s = "{}".format(f)
        if "e" in s or "E" in s:
            res = "{0:.{1}f}".format(f, decimal_places)
            return float(res)

        left, _, right = s.partition(".")

        if decimal_places == 0:
            return int(left)

        res = ".".join([left, (right + "0" * decimal_places)[:decimal_places]])
        return float(res)

    @staticmethod
    def decimal_places(n: int | float) -> int:
        """
        Returns the precision (number of decimal places) of a decimal number
        """
        if not isinstance(n, (float, int)):
            raise ValueError("The input must be a float or an integer")

        d = Decimal(f"{n}")
        return int(str(d.as_tuple().exponent).lstrip("-"))

    def compute_quote_qty(self, qty, market) -> float:
        raw_qty = qty * self.get_market_price(market=market)
        market_info = self.get_market_info(market)
        quote_qty = self.truncate_float(
            raw_qty, decimal_places=market_info["precision"]
        )
        return quote_qty

    @staticmethod
    def sort_by(array, key, descending=False, default=0):
        return sorted(
            array,
            key=lambda k: k[key] if k[key] is not None else default,
            reverse=descending,
        )

    @staticmethod
    def dt_to_unix(
        dt: datetime, tzinfo: timezone = timezone.utc, unit: str = "ms"
    ) -> int | None:
        unit_map = {"s": 1, "ms": 1e3, "ns": 1e6}
        if unit not in unit_map.keys():
            raise ValueError(
                f"Wrong unit {unit}. "
                f"Unit should be one of {list(unit_map.keys())}"
            )

        if dt is None:
            return None
        return int(dt.replace(tzinfo=tzinfo).timestamp() * unit_map[unit])

    @staticmethod
    def str_to_dt(
        string: str,
        format: str = "%Y-%m-%dT%H:%M:%SZ",
        tz: timezone = timezone.utc,
    ) -> datetime:
        return datetime.strptime(string, format).replace(tzinfo=tz)

    @staticmethod
    def _cast_values(params: dict[str, Any]) -> dict:
        return {}

    def prepare_params(self, params: dict[str, Any]) -> dict[str, Any]:
        return self._cast_values(self._clean_none_values(params))

    def handle_exception(self, r: requests.Response) -> None:
        return

    def request(
        self,
        http_method: str,
        url_path: str,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | str | None = None,
    ) -> Any:
        if not params:
            params = {}

        url = f"{self.base_url}{url_path}"

        if http_method == "GET":
            prep_params = self.prepare_params(params)
            r = self.session.request(http_method, url=url, params=prep_params)
        else:
            r = self.session.request(http_method, url=url, data=data)

        self.handle_exception(r)

        try:
            content = r.json()
        except JSONDecodeError:
            content = {"data": r.text}

        return content

    def signed_request(
        self,
        http_method: str,
        url_path: str,
        params: dict | None = None,
        data: dict | None = None,
    ) -> dict[str, Any]:
        return {}

    #################
    # EXCHANGE INFO #
    #################
    def get_server_time(self) -> Any:
        raise NotSupported(
            f"{self.name}: get_server_time() is not supported yet"
        )

    ###########
    # ACCOUNT #
    ###########
    def get_api_key_info(self) -> dict:
        raise NotSupported(
            f"{self.name}: get_api_key_info() is not supported yet"
        )

    def get_balance(self) -> dict:
        raise NotSupported(f"{self.name}: get_balance() is not supported yet")

    def get_transfer_history(
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        **kwargs,
    ) -> list[dict]:
        raise NotSupported(
            f"{self.name}: get_transfer_history() is not supported yet"
        )

    ###########
    # MARKETS #
    ###########
    def get_markets(self, **kwargs) -> list[dict]:
        raise NotSupported(f"{self.name}: get_markets() is not supported yet")

    def get_market_price(self, market: str) -> float:
        raise NotSupported(f"{self.name}: get_markets() is not supported yet")

    def get_market_info(self, market: str, **kwargs) -> dict[str, Any]:
        raise NotSupported(
            f"{self.name}: get_market_info() is not supported yet"
        )

    def get_ohlcv(
        self,
        market: str,
        timeframe: str = "1d",
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        **kwargs,
    ) -> list[list]:
        raise NotSupported(f"{self.name}: get_ohlcv() is not supported yet")

    #########
    # ORDER #
    #########
    def place_test_order(
        self,
        market: str,
        type: Literal["limit", "market"],
        side: Literal["buy", "sell"],
        qty: float,
        price: float | None = None,
        **kwargs,
    ):
        raise NotSupported(
            f"{self.name}: place_test_order() is not supported yet"
        )

    def place_order(
        self,
        market: str,
        type: Literal["limit", "market"],
        side: Literal["buy", "sell"],
        qty: float,
        price: float | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        raise NotSupported(f"{self.name}: place_order() is not supported yet")

    def get_open_orders(self, market: str | None = None) -> list[dict]:
        raise NotSupported(f"{self.name}: place_order() is not supported yet")

    def cancel_order(self, id: str) -> dict[str, Any]:
        raise NotSupported(f"{self.name}: cancel_order() is not supported yet")

    def cancel_orders(
        self, market: str | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        raise NotSupported(
            f"{self.name}: cancel_orders() is not supported yet"
        )



from typing import Literal

import pytest

from crypto_connector.base.errors import ExchangeError
from crypto_connector.base.exchange import Exchange
from crypto_connector.base.schemas import Order


@pytest.mark.parametrize(
    "exchange,market,type,side,qty,price",
    argvalues=[
        ("Binance", "EURUSDC", "limit", "buy", 8, 0.95),
        ("HTX", "USDCUSDT", "limit", "buy", 12, 0.95),
    ],
    indirect=["exchange"],
)
def test_place_order_ok(
    exchange: Exchange,
    market: str,
    type: Literal["limit", "market"],
    side: Literal["buy", "sell"],
    qty: float,
    price: float,
) -> None:
    order = exchange.place_order(
        market,
        type=type,
        side=side,
        qty=qty,
        price=price,
    )
    assert Order.model_validate(order)
    assert order.get("orderId") is not None

    # cancel order
    result = exchange.cancel_order(order["orderId"])
    assert result["success"] is True


@pytest.mark.parametrize(
    "exchange,market,type,side,qty,price",
    argvalues=[
        ("Binance", "EURUSDC", "limit", "buy", 1000, 0.95),
        ("HTX", "USDCUSDT", "limit", "buy", 1000, 0.95),
    ],
    indirect=["exchange"],
)
def test_insufficient_balance_order(
    exchange: Exchange,
    market: str,
    type: Literal["limit", "market"],
    side: Literal["buy", "sell"],
    qty: float,
    price: float,
) -> None:
    with pytest.raises(ExchangeError):
        exchange.place_order(market, type=type, side=side, qty=qty, price=price)


@pytest.mark.parametrize(
    "exchange,market,type,side,qty,price",
    argvalues=[
        ("Binance", "EURUSDC", "limit", "buy", 500, 0.95),
        ("HTX", "USDCUSDT", "limit", "buy", 500, 0.95),
    ],
    indirect=["exchange"],
)
def test_order_insufficient_balance(
    exchange: Exchange,
    market: str,
    type: Literal["limit", "market"],
    side: Literal["buy", "sell"],
    qty: float,
    price: float,
) -> None:
    with pytest.raises(ExchangeError):
        exchange.place_order(market, type=type, side=side, qty=qty, price=price)


@pytest.mark.parametrize(
    "exchange,market,type,side,qty,price",
    argvalues=[
        ("Binance", "ETHUSDC", "limit", "buy", 1e-09, 1000),
        ("HTX", "ETHUSDT", "limit", "buy", 1e-09, 1000),
    ],
    indirect=["exchange"],
)
def test_place_order_wrong_precision(
    exchange: Exchange,
    market: str,
    type: Literal["limit", "market"],
    side: Literal["buy", "sell"],
    qty: float,
    price: float,
) -> None:
    with pytest.raises(ExchangeError):
        exchange.place_order(market, type=type, side=side, qty=qty, price=price)


if __name__ == "__main__":
    pytest.main([__file__])



class BaseError(Exception):
    pass


class ExchangeError(BaseError):
    pass


class NotSupported(ExchangeError):
    pass


class BadResponse(ExchangeError):
    pass


class InvalidOrder(ExchangeError):
    pass


class MissingCredentials(ExchangeError):
    pass


class OrderNotFound(InvalidOrder):
    pass


class ServerError(BaseError):
    pass
